import java.util.ArrayList;
import java.util.HashSet;
import java.util.concurrent.*;

/**
 * PartialSolution class: Manages partial solutions to the puzzle. Used to find all valid solutions.
 */

public class PartialSolution{
    private static int intermediateSolutions;
    public static final int SIZE = 9;      // number of squares per row/column 
    public static final int BLOCKSIZE = 3; // number of squares per block row/column
  
    ArrayList<Row> rows;
  
    public PartialSolution() {
        intermediateSolutions++;
        rows = new ArrayList<Row>();
        for(int i=0; i<SIZE; i++) {
            rows.add(new Row());
        }
    }
  
  
    /**
     * Construct a new partial solution as a copy of the one given as a parameter
     */
    public PartialSolution(PartialSolution initial) {
        this();
        for(int i=0; i<SIZE; i++) {
            Row r = rows.get(i);
            Row initialRow = initial.rows.get(i);
            for (int j=0; j<SIZE; j++) {
                HashSet<Integer> cpy = new HashSet<Integer>();
                cpy.addAll(initialRow.elems().get(j));
                r.elems().set(j, cpy);
            }
        }
    }
  
    /**
     * Create a new partial solution based on input in a matrix
     */
    static PartialSolution fromMatrix(int[][] mat) {
        // do some basic error checking for the input matrix
        int len = mat.length;
        if (len != SIZE) {
            throw new RuntimeException("FromMatrix call received a matrix of size not equal with "+SIZE+"!");
        }
        for(int i =0; i<len; i++)
            if (mat[i].length != len) {
                throw new RuntimeException("FromMatrix call received jagged matrix as parameter!");
            }
    
        PartialSolution ps = new PartialSolution();
    
        // create the PartialSolution correrersponding to the input matrix
        // by calling setElement for of all non-zero elements in the input matrix
        for(int i=0; i<SIZE; i++)
            for(int j=0; j<SIZE; j++) {
                if (mat[i][j] != 0) {
                    ps = ps.setElement(i, j, mat[i][j]);
                }
            }
        return ps;
    }
  
  
    /**
     * Create a set of partial solutions taking into account all possible values in the square with coordinates (i, j)
     */
    public ArrayList<PartialSolution> selectSolutionsForElement(int i, int j) {
        ArrayList<PartialSolution> sols = new ArrayList<PartialSolution>();
    
        for(Integer val : rows.get(i).elems().get(j)) {
            PartialSolution p = this.setElement(i, j, val);
            sols.add(p);
        }
        return sols;
    }
  
    /** 
     * Gets the number of intermediate solutions used to solve the puzzle
     * return value undefined if called before findSolution
     */ 
    int getIntermediateSolutions()
    {
        return intermediateSolutions;
    }
  
    /**
     * Clears the puzzle
     */
    public void empty() {
        for(int i=0; i<SIZE; i++) {
            for(int j=0; j<SIZE; j++) 
                rows.get(i).elems().get(j).clear();
        }
    }
  
  
    /**
     * Sets the element with coordinates(row, col) to the value val. Applies constraints on the whole puzzle
     */
    public PartialSolution setElement(int row, int col, int val) {
        //do some basic error checking for the input value
        if ((val<1)||(val>SIZE)) 
            throw new RuntimeException("Tried to set element at "+row+" "+col+" at incorrect value "+val+"!");
    
        PartialSolution p = new PartialSolution(this);
        //addChild(p);
        HashSet<Integer> possibleValues = p.rows.get(row).elems().get(col);
        possibleValues.clear();
        possibleValues.add(val);
    
        // remove possible values from the same column
        // if they become impossible after setting the i,j element
        for(int ii=0; ii<SIZE; ii++)
            if (ii != row) {
                int initialSize =  p.rows.get(ii).elems().get(col).size();
                p.rows.get(ii).elems().get(col).remove((Integer)val);
                int finalSize =  p.rows.get(ii).elems().get(col).size();
                if ((initialSize == 2)&&(finalSize==1))
                    p = p.setElement(ii, col, p.rows.get(ii).elems().get(col).iterator().next()); 
            }
    
        // remove possible values from the same row
        // if they become impossible after setting the i,j element
        for(int jj=0; jj<SIZE; jj++)
            if (jj != col) {
                int initialSize =  p.rows.get(row).elems().get(jj).size();
                p.rows.get(row).elems().get(jj).remove((Integer)val);
                int finalSize =  p.rows.get(row).elems().get(jj).size();
                if ((initialSize == 2)&&(finalSize==1)) 
                    p = p.setElement(row, jj, p.rows.get(row).elems().get(jj).iterator().next());
            }
    
        // remove possible values from the same block
        // if they become impossible after setting the i,j element
        int block_i = row / BLOCKSIZE;
        int block_j = col / BLOCKSIZE;
        for(int ii=0; ii< SIZE/BLOCKSIZE; ii++)
            for(int jj=0; jj< SIZE/BLOCKSIZE; jj++) {
                if (! ((block_i*BLOCKSIZE + ii==row)&&(block_j*BLOCKSIZE+jj==col)) ) {
                    int initialSize =  p.rows.get(block_i*BLOCKSIZE+ii).elems().get(block_j*BLOCKSIZE+jj).size();
                    p.rows.get(block_i*BLOCKSIZE+ii).elems().get(block_j*BLOCKSIZE+jj).remove((Integer)val);
                    int finalSize =  p.rows.get(block_i*BLOCKSIZE+ii).elems().get(block_j*BLOCKSIZE+jj).size();
                    if ((initialSize == 2)&&(finalSize==1)) 
                        p = p.setElement(block_i*BLOCKSIZE+ii, block_j*BLOCKSIZE+jj, 
                                         p.rows.get(block_i*BLOCKSIZE+ii).elems().get(block_j*BLOCKSIZE+jj).iterator().next());
                }
            }
    
        return p;
    }
  
  
    /**
     * Solve the Sudoku puzzle
     */
    public ArrayList<PartialSolution> findSolution() {
        intermediateSolutions = 0;
        if (isFinal()) {
            ArrayList<PartialSolution> vSols = new ArrayList<PartialSolution>();
            vSols.add(this);
            return vSols;
        }
    
        return findSolution(0, 0);
    }
  
    /**
     * Enumerate different solutions starting from the square located at (initialI,initialJ)
     */
    private ArrayList<PartialSolution> findSolution(final int initialI, final int initialJ) {
        // Accumulate solutions in vSols
        ArrayList<PartialSolution> vSols = new ArrayList<PartialSolution>();
        if (rows.get(initialI).elems().get(initialJ).size()>1){
            // CASE 1: The square located at (initialI,initialJ) has > 1 possible values
            // Call selectSolutionsForElement to create a set of partial solutions, vs,
            // containing one for each possible value for the square located at (initialI,initialJ) 
            ArrayList<PartialSolution> vs = selectSolutionsForElement(initialI, initialJ);

            // Iterate through each solution in vs
            for(PartialSolution crtStepSol : vs){
                final PartialSolution stepSol = crtStepSol;
                /**** START OF CODE THAT CAN BE DECOMPOSED INTO A SEQUENTIAL/PARALLEL TASK ****/
                // Allocate vSolsLocal to capture subset of solutions starting from stepSol
                ArrayList<PartialSolution> vSolsLocal = new ArrayList<PartialSolution>();
                if (stepSol.isFinal()){
                    vSolsLocal.add(stepSol);
                }
                else if (!stepSol.isDeadEnd()){
                    if (initialJ<SIZE-1) vSolsLocal.addAll(stepSol.findSolution(initialI, initialJ+1));
                    else if (initialI<SIZE-1) vSolsLocal.addAll(stepSol.findSolution(initialI+1, 0));
                }
                /**** END OF CODE THAT CAN BE DECOMPOSED INTO A SEQUENTIAL/PARALLEL TASK ****/
                // Accumulate vSolsLocal into vSols
                vSols.addAll(vSolsLocal);
            }
        }
        else if (rows.get(initialI).elems().get(initialJ).size()==1) {
            // CASE 2: The square located at (initialI,initialJ) has exactly 1 value.
            // Recursively call findSolution() on the next square
            if (initialJ<SIZE-1){
                return findSolution(initialI, initialJ+1);
            }
            else if (initialI<SIZE-1){
                return findSolution(initialI+1, 0);
            }
        }
    
        return vSols;
    }
  
    /**
     * Checks if the partial solution has enough elements to lead to a correct solution
     */
    public boolean isDeadEnd() {
        for (int i=0; i<SIZE; i++)
            for(int j=0; j<SIZE; j++)
                if (rows.get(i).elems().get(j).size()==0) return true;
        return false;
    }
  
  
    /**
     * Checks if the solution is a final, correct solution of the game
     */
    public boolean isFinal() {
    
        // check that each square has only one element
        for (int i=0; i<SIZE; i++)
            for(int j=0; j<SIZE; j++)
                if (rows.get(i).elems().get(j).size()!=1) return false;
    
        // check that the values on each square are correct
        PartialSolution ps = new PartialSolution(this);
        for(int i=0; i<SIZE; i++)
            for(int j=0; j<SIZE; j++) 
                ps = ps.setElement(i, j, rows.get(i).elems().get(j).iterator().next() );
        if (ps == null || ps.isDeadEnd()) return false;
    
        return true;
    }
  
  
    /**
     * Creates a Sring representation of the puzzle
     */
    public String toString() {
        String s = "";
        int width = SIZE*3;
        for(int i=0; i< width; i++) 
            s += "-";
        s += System.getProperty("line.separator");
        for(Row r: rows) {
            s += "|" + r.toString()+"|"+System.getProperty("line.separator");
        }
        for(int i=0; i< width; i++) 
            s += "-";
        s+=System.getProperty("line.separator");
        return s;
    }
  
  
    /**
     * Checks if solution is correct
     */
    public boolean checkSolution(int[][] mat) {
    
        // check correspondence with initial matrix
        for(int i=0; i<SIZE; i++)
            for(int j=0; j<SIZE; j++) {
                if (mat[i][j] != 0) {
                    if ( (rows.get(i).elems().get(j).size()!=1) || 
                         (rows.get(i).elems().get(j).iterator().next() != mat[i][j]) ) {
                        return false;
                    }  
                }
            }
    
        // check that the current solution is a final, correct solution
        if (!isFinal()) return false;
    
        return true;
    }
}
